@@ -471,39 +471,49 @@ namespace ojph {
         converter = gen_cvrt_32b3c_to_16ub3c_be;
     }
 
-#ifndef OJPH_DISABLE_INTEL_SIMD
+#ifndef OJPH_DISABLE_SIMD
 
-    if (get_cpu_ext_level() >= X86_CPU_EXT_LEVEL_SSE41) {
-      if (bytes_per_sample == 1) {
-        if (num_components == 1) 
-          converter = sse41_cvrt_32b1c_to_8ub1c;
-        else
-          converter = sse41_cvrt_32b3c_to_8ub3c;
-      }
-      else {
-        if (num_components == 1) 
-          converter = sse41_cvrt_32b1c_to_16ub1c_be;
-        else
-          converter = sse41_cvrt_32b3c_to_16ub3c_be;
-      }
-    }
+  #if (defined(OJPH_ARCH_X86_64) || defined(OJPH_ARCH_I386))
 
-    if (get_cpu_ext_level() >= X86_CPU_EXT_LEVEL_AVX2) {
-      if (bytes_per_sample == 1) {
-        if (num_components == 1) 
-          converter = avx2_cvrt_32b1c_to_8ub1c;
-        else
-          converter = avx2_cvrt_32b3c_to_8ub3c;
+    #ifndef OJPH_DISABLE_SSE4
+      if (get_cpu_ext_level() >= X86_CPU_EXT_LEVEL_SSE41) {
+        if (bytes_per_sample == 1) {
+          if (num_components == 1) 
+            converter = sse41_cvrt_32b1c_to_8ub1c;
+          else
+            converter = sse41_cvrt_32b3c_to_8ub3c;
+        }
+        else {
+          if (num_components == 1) 
+            converter = sse41_cvrt_32b1c_to_16ub1c_be;
+          else
+            converter = sse41_cvrt_32b3c_to_16ub3c_be;
+        }
       }
-      else {
-        if (num_components == 1) 
-          converter = avx2_cvrt_32b1c_to_16ub1c_be;
-        else
-          { } // did not find an implementation better than sse41
-      }
-    }
+    #endif // !OJPH_DISABLE_SSE4
 
-#endif
+    #ifndef OJPH_DISABLE_AVX2
+      if (get_cpu_ext_level() >= X86_CPU_EXT_LEVEL_AVX2) {
+        if (bytes_per_sample == 1) {
+          if (num_components == 1) 
+            converter = avx2_cvrt_32b1c_to_8ub1c;
+          else
+            converter = avx2_cvrt_32b3c_to_8ub3c;
+        }
+        else {
+          if (num_components == 1) 
+            converter = avx2_cvrt_32b1c_to_16ub1c_be;
+          else
+            { } // did not find an implementation better than sse41
+        }
+      }
+    #endif // !OJPH_DISABLE_AVX2
+
+  #elif defined(OJPH_ARCH_ARM)
+
+  #endif // !(defined(OJPH_ARCH_X86_64) || defined(OJPH_ARCH_I386))
+
+#endif // !OJPH_DISABLE_SIMD
   }
 
   ////////////////////////////////////////////////////////////////////////////
