@@ -96,14 +96,25 @@ namespace ojph {
 
   ////////////////////////////////////////////////////////////////////////////
   enum OJPH_TILEPART_DIVISIONS: ui32 {
-    OJPH_TILEPART_NO_DIVISIONS  = 0x0, // no divisions to tile parts
-    OJPH_TILEPART_RESOLUTIONS = 0x1,
-    OJPH_TILEPART_COMPONENTS  = 0x2,
-    OJPH_TILEPART_LAYERS      = 0x4, // these are meaningless with HTJ2K
+    OJPH_TILEPART_NO_DIVISIONS = 0x0, // no divisions to tile parts
+    OJPH_TILEPART_RESOLUTIONS  = 0x1,
+    OJPH_TILEPART_COMPONENTS   = 0x2,
+    OJPH_TILEPART_LAYERS       = 0x4, // these are meaningless with HTJ2K
   };
 
   namespace local {
 
+    //defined here
+    struct param_siz;
+    struct param_cod;
+    struct param_qcd;
+    struct param_qcc;
+    struct param_cap;
+    struct param_sot;
+    struct param_tlm;
+    struct param_dfs;
+    struct param_atk;
+
     //////////////////////////////////////////////////////////////////////////
     enum JP2K_MARKER : ui16
     {
@@ -111,6 +122,7 @@ namespace ojph {
       CAP = 0xFF50, //extended capability
       SIZ = 0xFF51, //image and tile size (required)
       COD = 0xFF52, //coding style default (required)
+      COC = 0xFF53, //coding style component
       TLM = 0xFF55, //tile-part lengths
       PRF = 0xFF56, //profile
       PLM = 0xFF57, //packet length, main header
@@ -118,19 +130,20 @@ namespace ojph {
       CPF = 0xFF59, //corresponding profile values
       QCD = 0xFF5C, //qunatization default (required)
       QCC = 0xFF5D, //quantization component
-      COM = 0xFF64, //comment
-      SOT = 0xFF90, //start of tile-part
-      SOP = 0xFF91, //start of packet
-      EPH = 0xFF92, //end of packet
-      SOD = 0xFF93, //start of data
-      EOC = 0xFFD9, //end of codestream (required)
-
-      COC = 0xFF53, //coding style component
       RGN = 0xFF5E, //region of interest
       POC = 0xFF5F, //progression order change
       PPM = 0xFF60, //packed packet headers, main header
       PPT = 0xFF61, //packed packet headers, tile-part header
       CRG = 0xFF63, //component registration
+      COM = 0xFF64, //comment
+      DFS = 0xFF72, //downsampling factor styles
+      ADS = 0xFF73, //arbitrary decomposition styles
+      ATK = 0xFF79, //arbitrary transformation kernels
+      SOT = 0xFF90, //start of tile-part
+      SOP = 0xFF91, //start of packet
+      EPH = 0xFF92, //end of packet
+      SOD = 0xFF93, //start of data
+      EOC = 0xFFD9, //end of codestream (required)
     };
 
     //////////////////////////////////////////////////////////////////////////
@@ -189,8 +202,10 @@ namespace ojph {
         cptr[comp_num].YRsiz = (ui8)downsampling.y;
       }
 
-      void check_validity()
+      void check_validity(const param_cod& cod)
       {
+        this->cod = &cod;
+
         if (XTsiz == 0 && YTsiz == 0)
         { XTsiz = Xsiz + XOsiz; YTsiz = Ysiz + YOsiz; }
         if (Xsiz == 0 || Ysiz == 0 || XTsiz == 0 || YTsiz == 0)
@@ -224,10 +239,15 @@ namespace ojph {
       bool write(outfile_base *file);
       void read(infile_base *file);
 
+      void link(const param_cod* cod)
+      { this->cod = cod; }
+
+      void link(const param_dfs* dfs)
+      { this->dfs = dfs; }
+
       void set_skipped_resolutions(ui32 skipped_resolutions)
-      {
-        this->skipped_resolutions = skipped_resolutions;
-      }
+      { this->skipped_resolutions = skipped_resolutions; }
+      
       ui32 get_width(ui32 comp_num) const
       {
         assert(comp_num < get_num_components());
@@ -242,20 +262,16 @@ namespace ojph {
         ui32 t = ojph_div_ceil(Ysiz, ds) - ojph_div_ceil(YOsiz, ds);
         return t;
       }
+
+      point get_recon_downsampling(ui32 comp_num) const;
+      point get_recon_size(ui32 comp_num) const;
       ui32 get_recon_width(ui32 comp_num) const
-      {
-        assert(comp_num < get_num_components());
-        ui32 ds = (ui32)cptr[comp_num].XRsiz * (1u << skipped_resolutions);
-        ui32 t = ojph_div_ceil(Xsiz, ds) - ojph_div_ceil(XOsiz, ds);
-        return t;
-      }
+      { return get_recon_size(comp_num).x; }
       ui32 get_recon_height(ui32 comp_num) const
-      {
-        assert(comp_num < get_num_components());
-        ui32 ds = (ui32)cptr[comp_num].YRsiz * (1u << skipped_resolutions);
-        ui32 t = ojph_div_ceil(Ysiz, ds) - ojph_div_ceil(YOsiz, ds);
-        return t;
-      }
+      { return get_recon_size(comp_num).y; }
+
+      bool is_ws_kern_support_needed() { return ws_kern_support_needed; }
+      bool is_dfs_support_needed() { return dfs_support_needed; }
 
     private:
       ui16 Lsiz;
@@ -275,6 +291,10 @@ namespace ojph {
       ui32 skipped_resolutions;
       int old_Csiz;
       siz_comp_info store[4];
+      bool ws_kern_support_needed;
+      bool dfs_support_needed;
+      const param_cod* cod;
+      const param_dfs* dfs;
       param_siz(const param_siz&) = delete; //prevent copy constructor
       param_siz& operator=(const param_siz&) = delete; //prevent copy
     };
@@ -294,10 +314,18 @@ namespace ojph {
       ui8 block_style;
       ui8 wavelet_trans;
       ui8 precinct_size[33]; //num_decomp is in [0,32]
-    };
 
-    ///////////////////////////////////////////////////////////////////////////
-    typedef cod_SPcod cod_SPcoc;
+      size get_log_block_dims() const
+      { return size(block_width + 2, block_height + 2); }
+      size get_block_dims() const 
+      { size t = get_log_block_dims(); return size(1 << t.w, 1 << t.h); }
+      size get_log_precinct_size(ui32 res_num) const
+      {
+        assert(res_num <= num_decomp);
+        size ps(precinct_size[res_num] & 0xF, precinct_size[res_num] >> 4);
+        return ps;
+      }
+    };
 
     ///////////////////////////////////////////////////////////////////////////
     struct cod_SGcod
@@ -310,38 +338,66 @@ namespace ojph {
     ///////////////////////////////////////////////////////////////////////////
     struct param_cod
     {
+      // serves for both COD and COC markers
+
       friend ::ojph::param_cod;
+      ////////////////////////////////////////
       enum BLOCK_CODING_STYLES {
         VERT_CAUSAL_MODE = 0x8,
         HT_MODE = 0x40
       };
-    public:
+      ////////////////////////////////////////
+      enum cod_type : ui8 {
+        UNDEFINED = 0,
+        COD_MAIN  = 1,
+        COC_MAIN  = 2,
+        COD_TILE  = 3,
+        COC_TILE  = 4
+      };
+      ////////////////////////////////////////
+      enum dwt_type : ui8 {
+        DWT_IRV97 = 0,
+        DWT_REV53 = 1,
+      };
+
+    public: // COD_MAIN and COC_MAIN common functions
+      ////////////////////////////////////////
       param_cod()
       {
         memset(this, 0, sizeof(param_cod));
         SPcod.block_style = HT_MODE;
-        SGCod.prog_order = 2;
+        SGCod.prog_order = OJPH_PO_RPCL;
         SGCod.num_layers = 1;
         SGCod.mc_trans = 0;
         SPcod.num_decomp = 5;
         SPcod.block_width = 4; //64
         SPcod.block_height = 4; //64
-        set_reversible(false);
+        next = NULL;
       }
 
+      ////////////////////////////////////////
       void set_reversible(bool reversible)
       {
-        SPcod.wavelet_trans = reversible ? 1 : 0;
+        assert(type == UNDEFINED || type == COD_MAIN);
+        type = COD_MAIN;
+        SPcod.wavelet_trans = reversible ? DWT_REV53 : DWT_IRV97;
       }
 
+      ////////////////////////////////////////
       void employ_color_transform(ui8 val)
       {
         assert(val == 0 || val == 1);
+        assert(type == UNDEFINED || type == COD_MAIN);
+        type = COD_MAIN;
         SGCod.mc_trans = val;
       }
 
+      ////////////////////////////////////////
       void check_validity(const param_siz& siz)
       {
+        assert(type == UNDEFINED || type == COD_MAIN);
+        type = COD_MAIN;
+
         //check that colour transform and match number of components and
         // downsampling
         int num_comps = siz.get_num_components();
@@ -379,50 +435,131 @@ namespace ojph {
         }
       }
 
+      ////////////////////////////////////////
       ui8 get_num_decompositions() const
-      { return SPcod.num_decomp; }
-      size get_block_dims() const
       {
-        return size(1 << (SPcod.block_width + 2),
-                    1 << (SPcod.block_height + 2));
+        if (type == COD_MAIN)
+          return SPcod.num_decomp;
+        else if (type == COC_MAIN)
+        {
+          if (is_dfs_defined())
+            return parent->get_num_decompositions();
+          else
+            return SPcod.num_decomp;
+        }
+        else {
+          assert(0);
+          return 0; // just in case
+        }
       }
-      bool is_reversible() const
-      { return (SPcod.wavelet_trans == 1); }
+
+      ////////////////////////////////////////
+      size get_block_dims() const
+      { return SPcod.get_block_dims(); }
+
+      ////////////////////////////////////////
+      size get_log_block_dims() const
+      { return SPcod.get_log_block_dims(); }
+
+      ////////////////////////////////////////
+      ui8 get_wavelet_kern() const
+      { return SPcod.wavelet_trans; }
+
+      ////////////////////////////////////////
       bool is_employing_color_transform() const
       { return (SGCod.mc_trans == 1); }
-      size get_log_block_dims() const
-      { return size(SPcod.block_width + 2, SPcod.block_height + 2); }
+
+      ////////////////////////////////////////
       size get_precinct_size(ui32 res_num) const
       {
         size t = get_log_precinct_size(res_num);
-        t.w = 1 << t.w;
-        t.h = 1 << t.h;
-        return t;
+        return size(1 << t.w, 1 << t.h);
       }
+
+      ////////////////////////////////////////
       size get_log_precinct_size(ui32 res_num) const
-      {
-        assert(res_num <= SPcod.num_decomp);
-        size ps(15, 15);
+      { 
         if (Scod & 1)
-        {
-          ps.w = SPcod.precinct_size[res_num] & 0xF;
-          ps.h = SPcod.precinct_size[res_num] >> 4;
-        }
-        return ps;
+          return SPcod.get_log_precinct_size(res_num);
+        else
+          return size(15, 15);
       }
+
+      ////////////////////////////////////////
       bool packets_may_use_sop() const
-      { return (Scod & 2) == 2; }
+      { 
+        if (parent)
+          return (parent->Scod & 2) == 2; 
+        else
+          return (Scod & 2) == 2;
+      }
+
+      ////////////////////////////////////////
       bool packets_use_eph() const
-      { return (Scod & 4) == 4; }
+      { 
+        if (parent)
+          return (parent->Scod & 4) == 4;
+        else
+          return (Scod & 4) == 4;
+      }
 
+      ////////////////////////////////////////
       bool write(outfile_base *file);
-      void read(infile_base *file);
 
-    private:
-      ui16 Lcod;
-      ui8 Scod;
-      cod_SGcod SGCod;
-      cod_SPcod SPcod;
+      ////////////////////////////////////////
+      void read(infile_base *file, cod_type type);
+
+      ////////////////////////////////////////
+      void read(infile_base* file, cod_type type, ui32 num_comps, 
+                param_cod* cod);
+
+      ////////////////////////////////////////
+      void update_atk(const param_atk* atk);
+
+      ////////////////////////////////////////
+      void link_cod(const param_cod* coc)
+      { this->next = coc; }
+
+      ////////////////////////////////////////
+      const param_cod* get_cod(ui32 comp_num) const
+      {
+        const param_cod* result = this->next;
+        while (result != NULL && result->get_comp_num() != comp_num)
+          result = result->next;
+        if (result)
+          return result;
+        else
+          return this;
+      }
+
+      ////////////////////////////////////////
+      const param_atk* access_atk() const { return atk; }
+
+    public: // COC_MAIN only functions
+      ////////////////////////////////////////
+      bool is_dfs_defined() const 
+      { return (SPcod.num_decomp & 0x80) != 0; }
+
+      ////////////////////////////////////////
+      ui16 get_dfs_index() const  // cannot be more than 15
+      { return SPcod.num_decomp & 0xF; }
+
+      ////////////////////////////////////////
+      ui32 get_comp_num() const
+      { assert(type == COC_MAIN); return comp_num; }
+
+    private: // Common variables
+      cod_type type;        // The type of this cod structure
+      ui16 Lcod;            // serves as Lcod and Scod
+      ui8 Scod;             // serves as Scod and Scoc
+      cod_SGcod SGCod;      // Used in COD and copied to COC
+      cod_SPcod SPcod;      // serves as SPcod and SPcoc
+      const param_cod* next;// to chain coc parameters to cod
+
+    private: // COC only variables
+      param_cod* parent;    // parent COD structure
+      ui16 comp_num;        // component index of this COC structure
+      const param_atk* atk; // useful when SPcod.wavelet_trans > 1
     };
 
     ///////////////////////////////////////////////////////////////////////////
@@ -440,47 +577,55 @@ namespace ojph {
       { 
         Lqcd = 0;
         Sqcd = 0;
-        for (int i = 0; i < 97; ++i)
-          u16_SPqcd[i] = 0;
-        num_decomps = 0;
-        base_delta = -1.0f; 
+        memset(u16_SPqcd, 0, sizeof(u16_SPqcd));
+        num_subbands = 0;
+        base_delta = -1.0f;
       }
 
       void set_delta(float delta) { base_delta = delta; }
-      void set_rev_quant(ui32 bit_depth, bool is_employing_color_transform);
-      void set_irrev_quant();
 
       void check_validity(const param_siz& siz, const param_cod& cod)
       {
-        num_decomps = cod.get_num_decompositions();
-        if (cod.is_reversible())
+        ui32 num_decomps = cod.get_num_decompositions();
+        num_subbands = 1 + 3 * num_decomps;
+        if (cod.get_wavelet_kern() == param_cod::DWT_REV53)
         {
           ui32 bit_depth = 0;
           for (ui32 i = 0; i < siz.get_num_components(); ++i)
             bit_depth = ojph_max(bit_depth, siz.get_bit_depth(i));
-          set_rev_quant(bit_depth, cod.is_employing_color_transform());
+          set_rev_quant(num_decomps, bit_depth,
+            cod.is_employing_color_transform());
         }
-        else
+        else if (cod.get_wavelet_kern() == param_cod::DWT_IRV97)
         {
           if (base_delta == -1.0f) {
             ui32 bit_depth = 0;
             for (ui32 i = 0; i < siz.get_num_components(); ++i)
               bit_depth =
-                ojph_max(bit_depth, siz.get_bit_depth(i) + siz.is_signed(i));
+              ojph_max(bit_depth, siz.get_bit_depth(i) + siz.is_signed(i));
             base_delta = 1.0f / (float)(1 << bit_depth);
           }
-          set_irrev_quant();
-         }
+          set_irrev_quant(num_decomps);
+        }
+        else
+          assert(0);
       }
-
       ui32 get_num_guard_bits() const;
       ui32 get_MAGBp() const;
-      ui32 get_Kmax(ui32 resolution, ui32 subband) const;
-      float irrev_get_delta(ui32 resolution, ui32 subband) const;
+      ui32 get_Kmax(const param_dfs* dfs, ui32 num_decompositions,
+                    ui32 resolution, ui32 subband) const;
+      float irrev_get_delta(const param_dfs* dfs,
+                            ui32 num_decompositions,
+                            ui32 resolution, ui32 subband) const;
 
       bool write(outfile_base *file);
       void read(infile_base *file);
 
+    protected:
+      void set_rev_quant(ui32 num_decomps, ui32 bit_depth, 
+                         bool is_employing_color_transform);
+      void set_irrev_quant(ui32 num_decomps);
+
     protected:
       ui16 Lqcd;
       ui8 Sqcd;
@@ -489,8 +634,9 @@ namespace ojph {
         ui8 u8_SPqcd[97];
         ui16 u16_SPqcd[97];
       };
-      ui32 num_decomps;
-      float base_delta;
+      ui32 num_subbands;        // number of subbands
+      float base_delta;         // base quantization step size -- all other
+                                // step sizes are derived from it.
     };
 
     ///////////////////////////////////////////////////////////////////////////
@@ -502,7 +648,6 @@ namespace ojph {
     ///////////////////////////////////////////////////////////////////////////
     struct param_qcc : public param_qcd
     {
-      //friend ::ojph::param_qcc;
     public:
       param_qcc() : param_qcd()
       { comp_idx = 0; }
@@ -533,7 +678,7 @@ namespace ojph {
 
       void check_validity(const param_cod& cod, const param_qcd& qcd)
       {
-        if (cod.is_reversible())
+        if (cod.get_wavelet_kern() == param_cod::DWT_REV53)
           Ccap[0] &= 0xFFDF;
         else
           Ccap[0] |= 0x0020;
@@ -627,9 +772,138 @@ namespace ojph {
       Ttlm_Ptlm_pair* pairs;
       ui32 num_pairs;
       ui32 next_pair_index;
-      
     };
-  }
-}
+
+    ///////////////////////////////////////////////////////////////////////////
+    //
+    //
+    //
+    //
+    //
+    ///////////////////////////////////////////////////////////////////////////
+    struct param_dfs
+    {
+    public:
+      enum dfs_dwt_type : ui8 {
+        NO_DWT    = 0,  // no wavelet transform
+        BIDIR_DWT = 1,  // bidirectional DWT (this the conventional DWT)
+        HORZ_DWT  = 2,  // horizontal only DWT transform
+        VERT_DWT  = 3,  // vertical only DWT transform
+      };
+
+    public: // member functions
+      param_dfs() { memset(this, 0, sizeof(param_dfs)); }
+      ~param_dfs() { if (next) delete next; }
+      void init() { memset(this, 0, sizeof(param_dfs)); }
+      bool read(infile_base *file);
+      bool exists() const { return Ldfs != 0; }
+
+      // get_dfs return a dfs structure Sdfs == index, or NULL if not found
+      const param_dfs* get_dfs(int index) const;
+      // decomp_level is the decomposition level, starting from 1 for highest
+      // resolution to num_decomps for the coarsest resolution
+      dfs_dwt_type get_dwt_type(ui32 decomp_level) const;
+      ui32 get_subband_idx(ui32 num_decompositions, ui32 resolution,
+                           ui32 subband) const;
+      point get_res_downsamp(ui32 skipped_resolutions) const;
+
+    private: // member variables
+      ui16 Ldfs;       // length of the segment marker
+      ui16 Sdfs;       // index of this DFS marker segment
+      ui8 Ids;         // number of elements in Ddfs, 2 bits per sub-level
+      ui8 Ddfs[8];     // a string defining number of decomposition sub-levels
+                       // 8 bytes should be enough for 32 levels
+      param_dfs* next; // used for linking other dfs segments
+    };
+
+    ///////////////////////////////////////////////////////////////////////////
+    //
+    //
+    //
+    //
+    //
+    ///////////////////////////////////////////////////////////////////////////
+    // data structures used by param_atk
+
+    union lifting_step {
+      struct irv_data {
+        // si8 Oatk;     // only for arbitrary filter
+        // ui8 LCatk;    // number of lifting coefficients in a step
+        float Aatk;      // lifting coefficient
+      };
+
+      struct rev_data {
+        // si8 Oatk;     // only for arbitrary filter, offset of filter
+        ui8 Eatk;        // only for reversible, epsilon, the power of 2
+        si16 Batk;       // only for reversible, beta, the additive residue
+        // ui8 LCatk;    // number of lifting coefficients in a step
+        si16 Aatk;       // lifting coefficient
+      };
+
+      irv_data irv;
+      rev_data rev;
+    };
+
+    struct param_atk
+    {
+      // Limitations:
+      // Arbitrary filters (ARB) are not supported
+      // Up to 6 steps are supported -- more than 6 are not supported
+      // Only one coefficient per step -- first order filter
+      // Only even-indexed subsequence in first reconstruction step,
+      //   m_init = 0 is supported
+
+    public: // member functions
+      param_atk() { init(); }
+      ~param_atk() {
+        if (next && alloced_next) {
+          delete next;
+          next = NULL;
+        }
+        if (d != NULL && d != d_store) {
+          delete[] d;
+          init(false);
+        }
+      }
+      bool read(infile_base *file);
+      bool read_coefficient(infile_base *file, float &K);
+      bool read_coefficient(infile_base *file, si16 &K);
+      void init(bool clear_all = true) { 
+        if (clear_all)
+          memset(this, 0, sizeof(param_atk));
+        d = d_store; max_steps = sizeof(d_store) / sizeof(lifting_step);
+      }
+      void init_irv97();
+      void init_rev53();
+      void link(param_atk* next) 
+      { assert(this->next == NULL); this->next = next; alloced_next = false; }
+
+      ui8 get_index() const { return (ui8)(Satk & 0xFF); }
+      int get_coeff_type() const { return (Satk >> 8) & 0x7; }
+      bool is_whole_sample() const { return (Satk & 0x800) != 0; }
+      bool is_reversible() const { return (Satk & 0x1000) != 0; }
+      bool is_m_init0() const { return (Satk & 0x2000) == 0; }
+      bool is_using_ws_extension() const { return (Satk & 0x4000) != 0; }
+      const param_atk* get_atk(int index) const;
+      const lifting_step* get_step(ui32 s) const 
+      { assert(s < Natk); return d + s; }
+      ui32 get_num_steps() const { return Natk; }
+      float get_K() const { return Katk; }
+
+    private: // member variables
+      ui16 Latk;         // structure length
+      ui16 Satk;         // carries a variety of information
+      float Katk;        // only for irreversible scaling factor K
+      ui8 Natk;          // number of lifting steps
+      lifting_step* d;   // pointer to data, initialized to d_store
+      int max_steps;     // maximum number of steps without memory allocation
+      lifting_step d_store[6];   // lifting step coefficient
+      param_atk* next;   // used for chaining if more than one atk segment
+                         // exist in the codestream
+      bool alloced_next; // true if next was allocated, not just set to an
+                         // existing object
+    };
+  } // !local namespace
+} // !ojph namespace
 
 #endif // !OJPH_PARAMS_LOCAL_H
